from qiskit import Aer, IBMQ
from qiskit.providers.ibmq import least_busy
import numpy as np


class TestExecutor:

    def runTestAssertEqual(self,
                           qc,
                           nbTrials,
                           nbMeasurements,
                           qu0,
                           qu1,
                           measuredBit0,
                           measuredBit1,
                           basis,
                           backend):

        if basis.lower() != "z" and basis.lower() != "x" and basis.lower() != "y":
            raise Exception(f"unknown input basis: {basis}")

        if backend in ["aer_simulator", "aer_simulator_statevector", "aer_simulator_density_matrix", "aer_simulator_stabilizer", \
                "aer_simulator_matrix_product_state", "aer_simulator_extended_stabilizer", "aer_simulator_unitary", \
                "aer_simulator_superop", "qasm_simulator", "statevector_simulator", "unitary_simulator", "pulse_simulator"]:
            sim = Aer.get_backend(backend)
        elif backend.lower() == "ibmq":
            IBMQ.load_account()
            provider = IBMQ.get_provider(hub="ibm-q")
            available_backends = provider.backends(filters=lambda x: x.configuration().n_qubits >= len(qc.qubits) and \
                not x.configuration().simulator and x.status().operational==True) 

            if len(available_backends) == 0:
                raise Exception("No suitable quantum backend found")

            sim = least_busy(available_backends)
        else:
            sim = None


        if basis.lower() == "x":
            qc.h(qu0)
            qc.h(qu1)
        elif basis.lower() == "y":
            qc.sdg(qu0)
            qc.sdg(qu1)
            qc.h(qu0)
            qc.h(qu1)

        qc.measure(qu0, measuredBit0)
        qc.measure(qu1, measuredBit1)

        #print(qc)
        #qc.draw(output="mpl")
        #plt.show()

        trialProbas0 = np.empty(nbTrials)
        trialProbas1 = np.empty(nbTrials)

        for trialIndex in range(nbTrials):
            result = sim.run(qc, shots = nbMeasurements).result()
            counts = result.get_counts()
            #print(counts)
            #counts the results for each qubit
            nb0s_qu0 = nb0s_qu1 = 0
            for elem in counts:
                if elem[::-1][measuredBit0] == '0': nb0s_qu0 += counts[elem]
                if elem[::-1][measuredBit1] == '0': nb0s_qu1 += counts[elem]


            trialProba0 = nb0s_qu0 / nbMeasurements
            trialProba1 = nb0s_qu1 / nbMeasurements

            #print(trialproba0)
            #print(f"{nb0s_qu0}, {nbmeasurements - nb0s_qu1}")

            trialProbas0[trialIndex] = trialProba0
            trialProbas1[trialIndex] = trialProba1
        #print(trialprobas0)
        return (trialProbas0, trialProbas1)

    #Returns a list of list of tuple of tuple of int
    def runTestsAssertEqual(self,
                            initialisedTests,
                            nbTrials,
                            nbMeasurements,
                            qu0,
                            qu1,
                            measuredBit0,
                            measuredBit1,
                            basis,
                            backend):
        retList = []
        for qc in initialisedTests:
            retList.append(self.runTestAssertEqual(qc, nbTrials, nbMeasurements, qu0, qu1, measuredBit0, measuredBit1, basis, backend))
        return retList


    #Return for each test the recreated "statevector" of 2 bits
    def runTestAssertEntangled(self,
                               qc,
                               nbTrials,
                               nbMeasurements,
                               qu0,
                               qu1,
                               measuredBit0,
                               measuredBit1,
                               basis,
                               backend):

        if basis.lower() != "z" and basis.lower() != "x" and basis.lower() != "y":
            raise Exception(f"unknown input basis: {basis}")

        sim = Aer.get_backend(backend)

        if basis.lower() == "x":
            qc.h(qu0)
            qc.h(qu1)
        elif basis.lower() == "y":
            qc.sdg(qu0)
            qc.sdg(qu1)
            qc.h(qu0)
            qc.h(qu1)

        qc.measure(qu0, measuredBit0)
        qc.measure(qu1, measuredBit1)

        #print(qc)
        #qc.draw(output="mpl")
        #plt.show()

        trialVectors = np.zeros((nbTrials, 4))

        for trialIndex in range(nbTrials):
            result = sim.run(qc, shots = nbMeasurements).result()
            counts = result.get_counts()
            #print(counts)

            for key, value in counts.items():
                #Has to be reversed before
                if key[0:2] == "00":
                    trialVectors[trialIndex][0] = value / nbMeasurements
                #Measures for state |01>
                elif key[0:2] == "10":
                    trialVectors[trialIndex][1] = value / nbMeasurements
                #Measures for state |10> (no typo, the order is just reversed in get_counts)
                elif key[0:2] == "01":
                    trialVectors[trialIndex][2] = value / nbMeasurements
                elif key[0:2] == "11":
                    trialVectors[trialIndex][3] = value / nbMeasurements

        #print(trialVectors)
        return trialVectors



    def runTestsAssertEntangled(self,
                                initialisedTests,
                                nbTrials,
                                nbMeasurements,
                                qu0,
                                qu1,
                                measuredBit0,
                                measuredBit1,
                                basis,
                                backend):
        retList = []
        for qc in initialisedTests:
            retList.append(self.runTestAssertEntangled(qc, nbTrials, nbMeasurements, qu0, qu1, measuredBit0, measuredBit1, basis, backend))
        return retList



    def runTestAssertProbability(self,
                                 qc,
                                 nbTrials,
                                 nbMeasurements,
                                 qu0,
                                 measuredBit,
                                 basis,
                                 backend):

        if basis.lower() != "z" and basis.lower() != "x" and basis.lower() != "y":
            raise exception(f"unknown input basis: {basis}")

        sim = Aer.get_backend(backend)

        if basis.lower() == "x":
            qc.h(qu0)
        elif basis.lower() == "y":
            qc.sdg(qu0)
            qc.h(qu0)

        qc.measure(qu0, measuredBit)

        #print(qc)

        trialProbas = np.empty(nbTrials)

        for trialIndex in range(nbTrials):
            result = sim.run(qc, shots = nbMeasurements).result()
            counts = result.get_counts()
            #print(counts)
            #counts the results for each qubit
            nb0s = 0
            for elem in counts:
                #Bit oredering is in the reverse order for get_count
                #(if we measure the last bit, it will get its value in index 0 of the string for some reason)
                if elem[::-1][measuredBit] == '0': nb0s += counts[elem]


            trialProba = nb0s / nbMeasurements

            trialProbas[trialIndex] = trialProba
        #print(trialProbas)
        return trialProbas


    #Returns a tuple of tuple of bool
    #Data to collect is a tuple (one for each trial) of tuples (one for each trial) of bools (one for each measruement)
    def runTestsAssertProbability(self,
                                  initialisedTests,
                                  nbTrials,
                                  nbMeasurements,
                                  qu0,
                                  measuredBit,
                                  basis,
                                  backend):
        retList = []
        for qc in initialisedTests:
           retList.append(self.runTestAssertProbability(qc, nbTrials, nbMeasurements, qu0, measuredBit, basis, backend))
        return retList


    def runTestAssertTransformed(self,
                                 qc,
                                 nbTrials,
                                 nbMeasurements,
                                 qu0,
                                 backend):
        sim = Aer.get_backend(backend)

        qc = self.applyQuantumFunction(qc, inputProg, filename)

        #qc.draw(output="mpl")
        #plt.show()

        result = sim.run(qc, shots = nbMeasurements).result()
        #print('Statevecotor: ', result.get_statevector())
        return result.get_statevector()

    def runTestsAssertTransformed(self,
                                  tests,
                                  nbTrials,
                                  nbMeasurements,
                                  qu0,
                                  backend):
        retList = []
        for qc in tests:
            retList.append(runTestAssertTransformed(qc, nbTrials, nbMeasurements, qu0, backend))
        return retList



from math import degrees, pi
from .ExtraFunctions import cordsComplexToThetaPhi

class Qarg():
    #Look up for ways to make this cleaner
    def __init__(self,
                 arg0,
                 arg1 = None,
                 arg2 = None,
                 arg3 = None,
                 arg4 = None):
        if (isinstance(arg0, int) or isinstance(arg0, float)) \
            and (isinstance(arg1, int) or isinstance(arg1, float)) \
            and (isinstance(arg2, int) or isinstance(arg2, float)) \
            and (isinstance(arg3, int) or isinstance(arg3, float)) \
            and (arg4 == None or isinstance(arg4, bool)):

            if arg4 == None:
                arg4 = False

            self.init_theta_phi(arg0, arg1, arg2, arg3, arg4)

        elif (isinstance(arg0, list) or isinstance(arg0, tuple)) \
            and (isinstance(arg1, int) or isinstance(arg1, float) or arg1 == None) \
            and (isinstance(arg2, int) or isinstance(arg2, float) or arg2 == None) \
            and (isinstance(arg3, bool) or arg3 == None):
                if arg1 == None:
                    arg1 = 0
                if arg2 == None:
                    arg2 = 0
                if arg3 == None:
                    arg3 = False

                self.init_coords(arg0, arg1, arg2, arg3)
        elif (isinstance(arg0, str)):
            if arg0.lower() == "any":
                self.init_theta_phi(0, 360, 0, 360)
        else:
            raise Exception(f"Incorrect arguments for Qarg")

    def init_theta_phi(self,
                       minTheta,
                       maxTheta,
                       minPhi,
                       maxPhi,
                       radian):

        if not radian:
            #Error handling on Qarg values
            if minTheta < 0 or minTheta > 360:
                raise Exception(f"Invalid minTheta supplied, it has to be between 0 and 360 inclusive: {minTheta}")
            if maxTheta < 0 or maxTheta > 360:
                raise Exception(f"Invalid maxTheta supplied, it has to be between 0 and 360 inclusive: {maxTheta}")
            if minPhi < 0 or minPhi > 360:
                raise Exception(f"Invalid minPhi supplied, it has to be between 0 and 360 inclusive: {minPhi}")
            if maxPhi < 0 or maxPhi > 360:
                raise Exception(f"Invalid maxPhi supplied, it has to be between 0 and 360 inclusive: {maxPhi}")

            self.minTheta = minTheta
            self.maxTheta = maxTheta
            self.minPhi = minPhi
            self.maxPhi = maxPhi

        else:
            #Error handling is done through math.degrees()
            self.minTheta = degrees(minTheta)
            self.maxTheta = degrees(maxTheta)
            self.minPhi = degrees(minPhi)
            self.maxPhi = degrees(maxPhi)

    #Can initalise with 2 complex numbers and how much it can vary around
    def init_coords(self,
                    init_vect,
                    diff_theta,
                    diff_phi,
                    radian):

        (theta, phi) = cordsComplexToThetaPhi(init_vect)

        theta = degrees(theta)
        phi = degrees(phi)

        if not radian:
            self.minTheta = round(theta - diff_theta)
            self.maxTheta = round(theta + diff_theta)
            self.minPhi = round(phi - diff_phi)
            self.maxPhi = round(phi + diff_phi)
        else:
            self.minTheta = round(theta - degrees(diff_theta))
            self.maxTheta = round(theta + degrees(diff_theta))
            self.minPhi = round(phi - degrees(diff_phi))
            self.maxPhi = round(phi + degrees(diff_phi))




class TestProperty():
    def __init__(self,
                 p_value=0.01,
                 nbTests=10,
                 nbTrials=100,
                 nbMeasurements=500,
                 nbQubits=1,
                 nbClassicalBits=0,
                 qargs={}):

        #Error handling on all values for the test property
        if p_value < 0 and p_value > 1:
            raise Exception(f"Invalid p_value supplied: {p_value}")
        elif nbTests < 1:
            raise Exception(f"Invalid amount of tests supplied: {nbTests}")
        elif nbTrials < 1:
            raise Exception(f"Invalid number of trials supplied: {nbTrials}")
        elif nbMeasurements < 1:
            raise Exception(f"Invalid number of measurements supplied: {nbMeasurements}")

        elif isinstance(nbQubits, int) and nbQubits < 1:
            raise Exception(f"Invalid number of qubits supplied: {nbQubits}")
        elif ((isinstance(nbQubits, tuple) or isinstance(nbQubits, list))) \
                and (nbQubits[0] < 1 or nbQubits[1] < nbQubits[0] or len(nbQubits) != 2):
            raise Exception(f"Invalid range of qubits supplied: {nbQubits}")

        elif nbClassicalBits < 0:
            raise Exception(f"Invalid number of classical bits supplied: {nbClassicalBits}")

        for key, value in qargs.items():
            if key < 0 or key >= nbQubits:
                raise Exception(f"Invalid qubit index in the dictionary supplied: {(key, value)}")

        self.p_value = p_value
        self.nbTests = nbTests
        self.nbTrials = nbTrials
        self.nbMeasurements = nbMeasurements
        if isinstance(nbQubits, int):
            self.minQubits = nbQubits
            self.maxQubits = nbQubits
        elif isinstance(nbQubits, tuple) or isinstance(nbQubits, list):
            self.minQubits = nbQubits[0]
            self.maxQubits = nbQubits[1]
        self.nbClassicalBits = nbClassicalBits
        self.preconditions_q = qargs



from qiskit.circuit.quantumcircuit import QuantumCircuit
from .TestProperties import TestProperty
import random
from math import cos, sin, radians, degrees
import cmath
import numpy as np

class TestCaseGenerator:

    #Returns a tuple containing a QuantumCircuit, a list of the thetas used to initialise the qc and the phis used
    def generateTest(self, testProperties: TestProperty):

        qargs = testProperties.preconditions_q
        #cargs = testProperties.preconditions_c
        #TODO do cargs or remove?

        #Adds 2 classical bits to read the results of any assertion
        #Those bits will not interfere with the normal functioning of the program
        if testProperties.minQubits == testProperties.maxQubits:
            nbQubits = testProperties.minQubits
        else:
            nbQubits = np.random.randint(testProperties.minQubits, testProperties.maxQubits)

        qc = QuantumCircuit(nbQubits, testProperties.nbClassicalBits + 2)

        theta_init = {}
        phi_init = {}

        for key, value in qargs.items():
            #converts from degrees to radian
            randomTheta_deg = random.randint(value.minTheta, value.maxTheta)
            randomPhi_deg = random.randint(value.minPhi, value.maxPhi)

            theta_init[key] = randomTheta_deg
            phi_init[key] = randomPhi_deg

            randomTheta = radians(randomTheta_deg)
            randomPhi = radians(randomPhi_deg)

            value0 = cos(randomTheta/2)
            value1 = cmath.exp(randomPhi * 1j) * sin(randomTheta / 2)

            #print(f"values initialised: {degrees(randomTheta)}, {degrees(randomPhi)}")

            qc.initialize([value0, value1], key)

        return (qc, theta_init, phi_init)

    def generateTests(self, testProperties: TestProperty):
        return [self.generateTest(testProperties) for _ in range(testProperties.nbTests)]




from .TestProperties import TestProperty, Qarg
from .TestCaseGeneration import TestCaseGenerator
from .TestExecutionEngine import TestExecutor
from .StatisticalAnalysisEngine import StatAnalyser
from math import cos, radians

class QiskitPropertyTest():



    def assertEqualData(self, qu0, qu1, qu0_pre, qu1_pre, basis, filter_qc, backend):

        #Error handling for the parameters
        if not isinstance(qu0, int) or not isinstance(qu1, int) or not isinstance(qu0_pre, bool) or not isinstance(qu1_pre, bool):
            raise Exception(f"Incorrect arguments supplied to assertEqual: qu0: {qu0}, qu1: {qu1}; (Optional: qu0_pre: {qu0_pre}, qu1_pre: {qu1_pre}\nThere should be 2 integers, followed by 2 optional booleans, defaulted to False)")

        #TODO add more error handling

        generatedTests = [qc.copy() for qc, theta, phi in self.initialisedTests]

        numberedGeneratedTests = [x for x in enumerate(generatedTests)]

        filteredNumbered = list(filter(lambda x: len(list(filter(filter_qc, [x[1]]))) != 0, numberedGeneratedTests))

        indexNotFilteredOut = [x[0] for x in filteredNumbered]

        self.nbFiltered = len(indexNotFilteredOut)

        filteredTests = [x[1] for x in filteredNumbered]
        #print(f"DEBUG\nfilteredTests: {filteredTests}\nEND DEBUG\n")


        if qu0_pre == qu1_pre:

            #Applies the function to the generated tests only if they are both sampled after running the full program
            if not qu0_pre and not qu1_pre:
                for filteredTest in filteredTests:
                    self.quantumFunction(filteredTest)


            dataFromExec = TestExecutor().runTestsAssertEqual(filteredTests, self.testProperty.nbTrials, \
                self.testProperty.nbMeasurements, qu0, qu1, self.testProperty.nbClassicalBits, \
                self.testProperty.nbClassicalBits + 1, basis, backend)

            testResults = StatAnalyser().testAssertEqual(self.testProperty.p_value, dataFromExec)

            qu0Params = [(x[1].get(qu0, 0), x[2].get(qu0, 0)) for x in self.initialisedTests]
            qu1Params = [(x[1].get(qu1, 0), x[2].get(qu1, 0)) for x in self.initialisedTests]
            params = tuple(zip(qu0Params, qu1Params))

            testResultsWithInit = tuple(zip(testResults, params))

        else:
            generatedTestsPre = [qc.copy() for qc, theta, phi in self.initialisedTests]
            filteredTestsPre = [generatedTestsPre[x] for x in indexNotFilteredOut]

            for filteredTest in filteredTests:
                self.quantumFunction(filteredTest)

            if not qu0_pre:
                dataFrom_qu0 = TestExecutor().runTestsAssertProbability(generatedTests, self.testProperty.nbTrials, \
                    self.testProperty.nbMeasurements, qu0, self.testProperty.nbClassicalBits + 1, basis, backend)
            else:
                dataFrom_qu0 = TestExecutor().runTestsAssertProbability(filteredTestsPre, self.testProperty.nbTrials, \
                    self.testProperty.nbMeasurements, qu0, self.testProperty.nbClassicalBits + 1, basis, backend)

            if not qu1_pre:
                dataFrom_qu1 = TestExecutor().runTestsAssertProbability(generatedTests, self.testProperty.nbTrials, \
                    self.testProperty.nbMeasurements, qu1, self.testProperty.nbClassicalBits + 1, basis, backend)
            else:
                dataFrom_qu1 = TestExecutor().runTestsAssertProbability(filteredTestsPre, self.testProperty.nbTrials, \
                    self.testProperty.nbMeasurements, qu1, self.testProperty.nbClassicalBits + 1, basis, backend)


            formattedData = tuple(zip(dataFrom_qu0, dataFrom_qu1))
            testResults = StatAnalyser().testAssertEqual(self.testProperty.p_value, formattedData)

            qu0Params = [(x[1].get(qu0, 0), x[2].get(qu0, 0)) for x in self.initialisedTests]
            qu1Params = [(x[1].get(qu1, 0), x[2].get(qu1, 0)) for x in self.initialisedTests]
            params = tuple(zip(qu0Params, qu1Params))

            testResultsWithInit = tuple(zip(testResults, params))

        return testResultsWithInit




    def assertEqual(self, qu0, qu1, qu0_pre=False, qu1_pre=False, basis="z", filter_qc=lambda qc:True, backend="aer_simulator"):
        results = self.assertEqualData(qu0, qu1, qu0_pre, qu1_pre, basis, filter_qc, backend)

        if not qu0_pre and not qu1_pre:
            print(f"AssertEqual({qu0}, {qu1}) results using basis {basis.upper()}:")
        elif qu0_pre and qu1_pre:
            print(f"AssertEqual({qu0}_pre, {qu1}_pre) results using basis {basis.upper()}:")
        elif not qu0_pre and qu1_pre:
            print(f"AssertEqual({qu0}, {qu1}_pre) results using basis {basis.upper()}:")
        else:
            print(f"AssertEqual({qu0}_pre, {qu1}) results using basis {basis.upper()}:")

        failed = False
        nbFailed = 0
        for testIndex, testResult in enumerate(results):
            if not testResult[0]:
                failed = True
                nbFailed += 1
                print(f"Test at index {testIndex} failed with qubits initialised to:")
                print(f"qu0: Theta = {testResult[1][0][0]} Phi = {testResult[1][0][1]}")
                print(f"qu1: Theta = {testResult[1][1][0]} Phi = {testResult[1][1][1]}")

        if failed:
            print(f"Not all tests have succeeded: {self.nbFiltered - nbFailed} / {self.nbFiltered} succeeded\n")
        else:
            print(f"All {self.nbFiltered} tests have succeeded!\n")


        return results


    def assertNotEqual(self, qu0, qu1, qu0_pre=False, qu1_pre=False, basis="z", filter_qc=lambda qc:True, backend="aer_simulator"):
        oppositeResults = self.assertEqualData(qu0, qu1, qu0_pre, qu1_pre, basis, filter_qc, backend)

        results = [(not x[0], x[1]) for x in oppositeResults]

        if not qu0_pre and not qu1_pre:
            print(f"AssertNotEqual({qu0}, {qu1}) results using basis {basis.upper()}:")
        elif qu0_pre and qu1_pre:
            print(f"AssertNotEqual({qu0}_pre, {qu1}_pre) results using basis {basis.upper()}:")
        elif not qu0_pre and qu1_pre:
            print(f"AssertNotEqual({qu0}, {qu1}_pre) results using basis {basis.upper()}:")
        else:
            print(f"AssertNotEqual({qu0}_pre, {qu1}) results using basis {basis.upper()}:")

        failed = False
        nbFailed = 0
        for testIndex, testResult in enumerate(results):
            if not testResult[0]:
                failed = True
                nbFailed += 1
                print(f"Test at index {testIndex} failed with qubits initialised to:")
                print(f"qu0: Theta = {testResult[1][0][0]} Phi = {testResult[1][0][1]}")
                print(f"qu1: Theta = {testResult[1][1][0]} Phi = {testResult[1][1][1]}")

        if failed:
            print(f"Not all tests have succeeded: {self.nbFiltered - nbFailed} / {self.nbFiltered} succeeded\n")
        else:
            print(f"All {self.nbFiltered} tests have succeeded!\n")


        return results



    def assertEntangledData(self, qu0, qu1, basis, backend):

        generatedTests = [qc.copy() for qc, theta, phi in self.initialisedTests]

        for generatedTest in generatedTests:
            self.quantumFunction(generatedTest)

        dataFromExec = TestExecutor().runTestsAssertEntangled(generatedTests,
                                                              self.testProperty.nbTrials,
                                                              self.testProperty.nbMeasurements,
                                                              qu0,
                                                              qu1,
                                                              self.testProperty.nbClassicalBits,
                                                              self.testProperty.nbClassicalBits + 1,
                                                              basis,
                                                              backend)

        testResults = StatAnalyser().testAssertEntangled(self.testProperty.p_value, dataFromExec)


        qu0Params = [(x[1].get(qu0, 0), x[2].get(qu0, 0)) for x in self.initialisedTests]
        qu1Params = [(x[1].get(qu1, 0), x[2].get(qu1, 0)) for x in self.initialisedTests]
        params = tuple(zip(qu0Params, qu1Params))

        testResultsWithInit = tuple(zip(testResults, params))


        return testResultsWithInit


    def assertEntangled(self, qu0, qu1, basis="z", backend="aer_simulator"):
        results = self.assertEntangledData(qu0, qu1, basis, backend)

        print(f"AssertEntangled({qu0}, {qu1}) results using basis {basis.upper()}:")

        failed = False
        nbFailed = 0
        for testIndex, testResult in enumerate(results):
            if not testResult[0]:
                failed = True
                nbFailed += 1
                print(f"Test at index {testIndex} failed with qubits initialised to:")
                print(f"qu0: Theta = {testResult[1][0][0]} Phi = {testResult[1][0][1]}")
                print(f"qu1: Theta = {testResult[1][1][0]} Phi = {testResult[1][1][1]}")

        if failed:
            print(f"Not all tests have succeeded: {len(self.initialisedTests) - nbFailed} / {len(self.initialisedTests)} succeeded\n")
        else:
            print(f"All {len(self.initialisedTests)} tests have succeeded!\n")


        return results



    def assertNotEntangled(self, qu0, qu1, basis="z", backend="aer_simulator"):
        oppositeResults = self.assertEntangledData(qu0, qu1, basis, backend)

        results = [(not x[0], x[1]) for x in oppositeResults]

        print(f"AsserNotEntangled({qu0}, {qu1}) results using basis {basis.upper()}:")

        failed = False
        nbFailed = 0
        for testIndex, testResult in enumerate(results):
            if not testResult[0]:
                failed = True
                nbFailed += 1
                print(f"Test at index {testIndex} failed with qubits initialised to:")
                print(f"qu0: Theta = {testResult[1][0][0]} Phi = {testResult[1][0][1]}")
                print(f"qu1: Theta = {testResult[1][1][0]} Phi = {testResult[1][1][1]}")

        if failed:
            print(f"Not all tests have succeeded: {len(self.initialisedTests) - nbFailed} / {len(self.initialisedTests)} succeeded\n")
        else:
            print(f"All {len(self.initialisedTests)} tests have succeeded!\n")


        return results




    def assertProbabilityData(self, qu0, expectedProba, qu0_pre, basis, filter_qc, backend):

        #Error handling
        if not isinstance(qu0, int) or not (isinstance(expectedProba, float) \
           or isinstance(expectedProba, int)) or not isinstance(qu0_pre, bool):
            print(f"qu0: {qu0}, expectedProba: {expectedProba}; (Optional: qu0_pre: {qu0_pre}")
            print(f"There should be an integer, followed by either a float or an int, followed by an optional boolean")
            raise Exception(f"Incorrect arguments supplied to assertProbability")

        expectedProbas = [expectedProba for _ in range(self.testProperty.nbTests)]

        generatedTests = [qc.copy() for qc, theta, phi in self.initialisedTests]

        #Only apply the functions if specified
        if not qu0_pre:
            for generatedTest in generatedTests:
                self.quantumFunction(generatedTest)

        filteredTests = list(filter(filter_qc, generatedTests))

        self.nbFiltered = len(filteredTests)



        dataFromExec = TestExecutor().runTestsAssertProbability(filteredTests,
                                                                self.testProperty.nbTrials,
                                                                self.testProperty.nbMeasurements,
                                                                qu0,
                                                                self.testProperty.nbClassicalBits + 1,
                                                                basis,
                                                                backend)

        testResults = StatAnalyser().testAssertProbability(self.testProperty.p_value, expectedProbas, dataFromExec)

        qu0Params = [(x[1].get(qu0, 0), x[2].get(qu0, 0)) for x in self.initialisedTests]

        testResultsWithInit = tuple(zip(testResults, qu0Params))



        return testResultsWithInit



    def assertProbability(self, qu0, expectedProba, qu0_pre=False, basis="z", filter_qc=lambda qc:True, backend="aer_simulator"):
        results = self.assertProbabilityData(qu0, expectedProba, qu0_pre, basis, filter_qc, backend)

        if not qu0_pre:
            print(f"AssertProbability({qu0}, {expectedProba}) results using basis {basis.upper()}:")
        else:
            print(f"AssertProbability({qu0}_pre, {expectedProba}) results using basis {basis.upper()}:")

        failed = False
        nbFailed = 0
        for testIndex, testResult in enumerate(results):
            if not testResult[0]:
                failed = True
                nbFailed += 1
                print(f"Test at index {testIndex} failed with qubits initialised to:")
                print(f"qu0: Theta = {testResult[1][0]} Phi = {testResult[1][1]}")

        if failed:
            print(f"Not all tests have succeeded: {self.nbFiltered - nbFailed} / {self.nbFiltered} succeeded\n")
        else:
            print(f"All {self.nbFiltered} tests have succeeded!\n")


        return results




    def assertNotProbability(self, qu0, expectedProba, qu0_pre=False, basis="z", filter_qc=lambda qc:True, backend="aer_simulator"):
        oppositeResults = self.assertProbabilityData(qu0, expectedProba, qu0_pre, basis, filter_qc, backend)

        results = [(not x[0], x[1]) for x in oppositeResults]

        if not qu0_pre:
            print(f"AssertNotProbability({qu0}, {expectedProba}) results using basis {basis.upper()}:")
        else:
            print(f"AssertNotProbability({qu0}_pre, {expectedProba}) results using basis {basis.upper()}:")

        failed = False
        nbFailed = 0
        for testIndex, testResult in enumerate(results):
            if not testResult[0]:
                failed = True
                nbFailed += 1
                print(f"Test at index {testIndex} failed with qubits initialised to:")
                print(f"qu0: Theta = {testResult[1][0]} Phi = {testResult[1][1]}")

        if failed:
            print(f"Not all tests have succeeded: {len(self.initialisedTests) - nbFailed} / {len(self.initialisedTests)} succeeded\n")
        else:
            print(f"All {len(self.initialisedTests)} tests have succeeded!\n")


        return results




    def assertTeleportedData(self, sent, received, basis, backend):

        generatedTests = [qc.copy() for qc, theta, phi in self.initialisedTests]

        for generatedTest in generatedTests:
            self.quantumFunction(generatedTest)


        expectedProbas = []
        for qc, thetas, phis in self.initialisedTests:
            expectedProba = cos(radians(thetas[sent]) / 2) ** 2
            expectedProbas.append(expectedProba)

        dataFromReceived = TestExecutor().runTestsAssertProbability(generatedTests,
                                                                    self.testProperty.nbTrials,
                                                                    self.testProperty.nbMeasurements,
                                                                    received,
                                                                    self.testProperty.nbClassicalBits + 1,
                                                                    basis,
                                                                    backend)

        testResults = StatAnalyser().testAssertProbability(self.testProperty.p_value, expectedProbas, dataFromReceived)

        qu0Params = [(x[1].get(sent, 0), x[2].get(sent, 0)) for x in self.initialisedTests]
        qu1Params = [(x[1].get(received, 0), x[2].get(received, 0)) for x in self.initialisedTests]
        params = tuple(zip(qu0Params, qu1Params))

        testResultsWithInit = tuple(zip(testResults, params))


        return testResultsWithInit


    def assertTeleported(self, sent, received, basis="z", backend="aer_simulator"):
        results = self.assertTeleportedData(sent, received, basis, backend)

        print(f"AssertTeleported({sent}, {received}) results using basis {basis.upper()}:")

        failed = False
        nbFailed = 0
        for testIndex, testResult in enumerate(results):
            if not testResult[0]:
                failed = True
                nbFailed += 1
                print(f"Test at index {testIndex} failed with qubits initialised to:")
                print(f"sent: Theta = {testResult[1][0][0]} Phi = {testResult[1][0][1]}")

        if failed:
            print(f"Not all tests have succeeded: {len(self.initialisedTests) - nbFailed} / {len(self.initialisedTests)} succeeded\n")
        else:
            print(f"All {len(self.initialisedTests)} tests have succeeded!\n")


        return results






    def run(self):
        print(f"Running tests for {type(self).__name__}:\n")

        self.testProperty = self.property()

        self.initialisedTests = TestCaseGenerator().generateTests(self.testProperty)

        self.assertions()

        print(f"Tests for {type(self).__name__} finished\n")

    def runTests(self):
        self.run()



from math import acos, sin, isclose, pi

def cordsComplexToThetaPhi(coords):
    #coords is a tuple/list of 2 complex numbers
    if not len(coords) == 2:
        raise Exception(f"Length of input is not 2: {coords}")
    #if not isinstance(coords[0], complex) and not isinstance(coords[1], complex):
    #    raise Exception(f"Inputs are not 2 complex numbers: {coords}")

    magnitude = abs(coords[0]) ** 2 + abs(coords[1]) ** 2
    #print(f"values: {coords}")
    #print(f"coords[0]: {coords[0]}")
    #print(f"coords[1]: {coords[1]}")

    if not isclose(magnitude.real, 1):
        raise Exception(f"Magnitude of vector not 1. Make sure that the sum of both squares is 1. Magnitude: {magnitude}")


    if isclose(coords[0].real, 1):
        theta = 0
        phi = 0
    elif isclose(coords[0].real, 0):
        theta = pi
        phi = 0
    else:
        theta = acos(coords[0].real) * 2
        #print((coords[1].real) / sin(theta / 2))
        phi = acos((coords[1].real) / sin(theta / 2))

    return (theta, phi)



from setuptools import setup, find_packages

setup(
    name="qiskit-ptesting",
    version="0.1.0.0.9",
    license="Apache 2.0",
    author="Pierre Brassart",
    author_email="pierrebrassart80@hotmail.fr",
    packages=find_packages("src"),
    package_dir={"": "src"},
    url="https://gitlab.com/twistercool/Qiskit-PTesting",
    classifiers=[
        "Development Status :: 3 - Alpha",
        ],
    keywords="qiskit property testing",
    install_requires=[
            "qiskit",
            "scipy",
            "numpy"
        ],
    python_requires=">=3.6",
    long_description="README",
    long_description_content_type='text/markdown',
)



from math import acos, sin, degrees, degrees, isclose
from scipy import stats
from qiskit.quantum_info import entanglement_of_formation
import numpy as np

class StatAnalyser:

    #use 2-sample t-test
    #null hypothesis is that the two states are equal
    def testAssertEqual(self,
                        p_value,
                        data):
        testResults = []
        for testIndex in range(len(data)):
            if np.array_equal(data[testIndex][0], data[testIndex][1]):
                testResults.append(True)
                continue

            t, stat_p = stats.ttest_ind(data[testIndex][0], data[testIndex][1])
            testResults.append(p_value <= stat_p)

        return testResults


    def testAssertEntangled(self,
                            p_value,
                            data):
        #print(testResults)
        testResults = []
        for testIndex in range(len(data)):
            totalEntanglement = 0
            for trialResult in data[testIndex]:
                #print(np.sqrt(trialResult))
                totalEntanglement += entanglement_of_formation(np.sqrt(trialResult))
                #print(entanglement_of_formation(np.sqrt(trialResult)))
            totalEntanglement /= len(data[testIndex])
            #print(f"total entanglement: {totalEntanglement}")
            testResults.append(totalEntanglement >= 1 - p_value)

        return testResults



    #Using one-sample t-test
    #The Null hypothesis is that the sample was taken with the same probability as the argument
    #TestResults is a tuple of tuple of bool
    def testAssertProbability(self,
                            p_value,
                            expectedProbas,
                            data):

        testResults = []

        for trialIndex in range(len(data)):
            if np.all(data[trialIndex] == data[trialIndex][0]):
                testResults.append(isclose(data[trialIndex][0], expectedProbas[trialIndex], abs_tol=0.01))
                continue

            t, stat_p = stats.ttest_1samp(data[trialIndex], expectedProbas[trialIndex])

            testResults.append(p_value <= stat_p)

            #print(stat_p)

        return testResults



    def testAssertTransformed(self, thetaMin, thetaMax, phiMin, phiMax, testStatevectors):

        results = []
        for statevector in testStatevectors:

            realPart0 = float(statevector[0].real)

            thetaRad = acos(realPart0) * 2

            #print(f"thata degrees from statevector: {degrees(thetaRad)}")

            if sin(thetaRad / 2) == 0:
                raise Exception("Division by 0")

            phiRad1 = acos(float(statevector[1].real) / sin(thetaRad / 2))
            #phiRad2 = asin(float(testStatevector[1].imag) / sin(thetaRad / 2))

            #print(f"values from statevectors: {degrees(thetaRad)}, {degrees(phiRad1)}")

            results.append(degrees(thetaRad) >= thetaMin - 0.001 and degrees(thetaRad) <= thetaMax + 0.001 and \
                           degrees(phiRad1) >= phiMin - 0.001 and degrees(phiRad1) <= phiMax + 0.001)

        return results
